#include <l4/sys/ipc.h>

#include "config.h"
#include "kdb.h"
#include "thread.h"
#include "thread_list.h"
#include "irq.h"
#include "time.h"

#include "ipc.h"

// the rendezvous signal handler: sent by the IPC sender to the receiver
unsigned 
thread_t::ipc_send_regs(const thread_user_regs_t *sender_regs)
{
  if (! in_transaction())
    {
      unsigned ret = trans_invoke(this, &ipc_send_regs, sender_regs);

      // if send_signal() has to abort for some reason, convert to an IPC
      // error code here
      return (ret == static_cast<unsigned long>(signal_t::Sig_aborted))
	? L4_IPC_SECANCELED
	: ret;
    }

  if (state() == Thread_invalid)
    {
      _current_trans->sender()->sender_dequeue(this);
      return L4_IPC_ENOT_EXISTENT;
    }

  unsigned ipc_state = state() & (Thread_receiving|Thread_waiting
				  | Thread_send_in_progress
				  | Thread_ipc_in_progress);

  // Thread_send_in_progress must not be set
  if (! (ipc_state == (Thread_waiting | Thread_ipc_in_progress))
      && (! ((ipc_state == (Thread_receiving | Thread_ipc_in_progress)) 
	     && my_partner == _current_trans->sender())))
    return 0x80000000;		// transient error -- partner not ready

  if (sender_regs->eax & 1)
    panic("deceiving ipc");	// XXX unimplemented

  unsigned long ret;

  ret = my_receive_regs->eax = 0; // error code: IPC successful
  // dequeue from sender queue if enqueued
  _current_trans->sender()->sender_dequeue(this); 

  // Reset sender's timeout, if any.  Once we're here, we don't want
  // the timeout to reset the sender's ipc_in_progress state bit
  // (which it still needs for a subsequent receive operation).
  if (timeout_t *t = _current_trans->sender()->my_timeout)
    {
      t->reset();
      if (t->has_hit())		// too late?
	_current_trans->sender()->state_add(Thread_ipc_in_progress); // fix
    }

  // copy message register contents
  my_receive_regs->edx = sender_regs->edx;
  my_receive_regs->ebx = sender_regs->ebx;

  // copy sender ID
  *my_receive_regs->thread_esi_edi() = _current_trans->sender()->id();

  // is this a fast (registers-only) receive operation?
  // the following operations can be short-cut here:
  // - long or message sender to short message receiver
  // - (short) register message sender to long message receiver
  // the following operations can not be short-cut as we're not allowed 
  // to touch :
  // - long message sender to long message receiver (obviously)
  // - short-flexpage message server to long message receiver 
  //   (because the rcvr's msg buffer may contain a flexpage option)
  if ((my_receive_regs->ebp & ~1) == 0 // receiver wants register IPC
      || (my_receive_regs->ebp & 2) // receiver wants short flexpage
      || (sender_regs->eax & ~1) == 0) // sender wants register IPC
    {
      // short IPC!

      if (sender_regs->eax & ~3) // long message sender?
	{
	  my_receive_regs->eax = L4_IPC_REMSGCUT;
	  ret = L4_IPC_SEMSGCUT;
	}
      else if (sender_regs->eax & 2) // send short flexpage?
	{
	  if (! (my_receive_regs->ebp & 2)) // rcvr not expecting an fpage?
	    {
	      my_receive_regs->eax = L4_IPC_REMSGCUT;
	      ret = L4_IPC_SEMSGCUT;
	    }
	  else
	    {
	      my_receive_regs->eax = 
		fpage_map(_current_trans->sender()->space(), 
			  (l4_fpage_t){fpage: sender_regs->ebx},
			  space(),
			  (l4_fpage_t){fpage: my_receive_regs->ebp},
			  sender_regs->edx & ~PAGE_MASK);
	  
	      if ((my_receive_regs->eax & L4_IPC_ERROR_MASK)
		  == L4_IPC_REMAPFAILED)
		ret = L4_IPC_SEMAPFAILED;
	    }
	}
 
      // (If we don't send a flexpage in our message, that's OK even
      // if the receiver expected one.  The receiver can tell from the
      // status code that he didn't receive one.)

      // IPC done -- reset states
      _current_trans->sender()->state_del(Thread_polling 
					  | Thread_send_in_progress);
      
      state_change(~(Thread_waiting | Thread_receiving | Thread_busy
		     | Thread_ipc_in_progress),
		   Thread_running);
    }
  else
    {
      // prepare long IPC

      // XXX check for cancel -- sender?  receiver?

      my_partner = _current_trans->sender();
      my_pagein_request = 0;
      _current_trans->sender()->my_target_desc = my_receive_regs->ebp & ~3;

      // if the receive timeout has hit, ignore it.  we do this by
      // overwriting the Thread_ipc_in_process flag which the timeout
      // may have deleted.

      if (my_timeout)
	my_timeout->reset();
      
      // set up return code in case we're aborted
      my_receive_regs->eax = L4_IPC_REABORTED;

      state_change(~(Thread_waiting | Thread_receiving), 
		   Thread_busy | Thread_ipc_in_progress);

      // in the sender, send_in_progress must remain set
      _current_trans->sender()->state_del(Thread_polling);
    }

  _current_trans->set_switch_hint(signal_t::Ret_activate_receiver);

  return ret;
}

// the "receiver is ready" signal handler: sent by IPC receiver to sender
unsigned
thread_t::ipc_receiver_ready()
{
  if (! in_transaction())	// not already in signal?
    {
      return trans_invoke(this, &ipc_receiver_ready);
    }
  
  if (! state_del(Thread_polling))
    return signal_t::Sig_fail;

  state_add(Thread_running);

  _current_trans->set_switch_hint(signal_t::Ret_activate_receiver);

  return signal_t::Sig_success;
}

// the send operation
unsigned 
thread_t::do_send(thread_t *partner, l4_timeout_t t, thread_user_regs_t *regs)
{
  if (!partner || partner->state() == Thread_invalid // partner nonexistent?
      || partner == nil_thread) // special case: must not send to L4_NIL_ID
    return L4_IPC_ENOT_EXISTENT;
      
  unsigned ret;

  // register partner so that we can be dequeued by kill()
  my_send_partner = partner;	

  state_add(Thread_polling | Thread_ipc_in_progress 
	    | Thread_send_in_progress);
  sender_enqueue(partner);
  
  // try a rendezvous with the partner
  ret = partner->ipc_send_regs(regs);
      
  if (ret & 0x80000000)		// transient error
    {
      // in case of a transient error, try sleeping
      
      timeout_t timeout;
      
      if (t.to.snd_exp != 0) // not sleep forever?
	{
	  if (t.to.snd_man == 0) // timeout = zero?
	    {
	      state_del(Thread_polling | Thread_ipc_in_progress
			| Thread_send_in_progress);
	      sender_dequeue(partner);
	      return L4_IPC_SETIMEOUT; // give up
	    }      
	  
	  // enqueue a timeout
	  my_timeout = &timeout;
	  timeout.set(kmem::info()->clock 
		      + timeout_to_microsec(t.to.snd_man, t.to.snd_exp));
	  
	}
      
      do 
	{
	  // detect a reset thread here (before scheduling)
	  if (state() & Thread_cancel)
	    {
	      // we've presumably been reset!
	      sender_dequeue(partner);
	      ret = L4_IPC_SECANCELED;
	      break;
	    }

	  // go to sleep
	  if (state_change_safely(~(Thread_ipc_in_progress
				    | Thread_polling | Thread_running), 
				  Thread_ipc_in_progress | Thread_polling))
	    schedule();
	  
	  state_add(Thread_polling);
	  
	  // detect if we have been reset meanwhile
	  if (state() & Thread_cancel)
	    {
	      // we've presumably been reset!
	      sender_dequeue(partner);
	      ret = L4_IPC_SECANCELED;
	      break;
	    }
	  
	  // detect if we have timed out
	  if (timeout.has_hit())
	    {
	      sender_dequeue(partner);
	      ret = L4_IPC_SETIMEOUT;
	      break;
	    }
	  
	  // OK, it's our turn to try again to send the message
	  ret = partner->ipc_send_regs(regs);	      
	}
      while ((ret & 0x80000000)); // transient error
				// re-enter loop even if !(state&polling)
				// for error handling
      assert(! in_sender_list());
  
      if (timeout.is_set())
	timeout.reset();

      my_timeout = 0;
    }

  assert(! in_sender_list());

  // are we in the middle of a long IPC?
  if (ret == 0 && (state() & Thread_send_in_progress))
    {
      // long IPC
      ret = do_send_long(partner, t, regs);
    }

  state_del(Thread_polling | Thread_send_in_progress
	    | ((state() & (Thread_receiving | Thread_waiting | Thread_busy))
	       ? 0 : Thread_ipc_in_progress));

  return ret;      
}

// preparation for the receive operation
void 
thread_t::prepare_receive(sender_t *partner,
			  thread_user_regs_t *regs)
{
  if (partner && partner->state() == Thread_invalid) // partner nonexistent?
    return;			// just return
  // don't need to signal error here -- it will be detected later

  unsigned ipc_state;

  my_receive_regs = regs;	// message should be poked in here

  if (partner)
    {
      my_partner = partner;
      ipc_state = Thread_receiving;
    }
  else
    ipc_state = Thread_waiting;

  // if we don't have a send, get us going here; otherwise, the send
  // operation will set the Thread_ipc_in_progress flag after it has
  // initialized
  if (regs->eax == 0xffffffff)
    state_add(ipc_state | Thread_ipc_in_progress);
  else
    state_add(ipc_state);
}

// the receive operation
unsigned 
thread_t::do_receive(sender_t *partner, l4_timeout_t t, 
		     thread_user_regs_t *regs)
{
  unsigned ipc_state = partner ? Thread_receiving : Thread_waiting;

  timeout_t timeout;

  while ((state() & (ipc_state | Thread_ipc_in_progress | Thread_cancel))
	 == (ipc_state | Thread_ipc_in_progress))
    {
      state_add(Thread_busy);

      // if a sender has been queued, wake it up
      sender_t *first = sender_first; // find a sender
      if (first)
	{
	  if (!partner || partner == first) // first is acceptible sender
	    first->ipc_receiver_ready(); // wake it up

	  else if (partner->in_sender_list() // partner in list, but not first?
		   && this == partner->my_send_partner)
	    partner->ipc_receiver_ready();
	}
      
      if (! my_timeout && t.to.rcv_exp != 0)
	{
	  my_timeout = &timeout;
      
	  timeout.set(kmem::info()->clock 
		      + timeout_to_microsec(t.to.rcv_man, t.to.rcv_exp));
	  
	}

      if (! state_change_safely(~(Thread_ipc_in_progress | ipc_state 
				  | Thread_running | Thread_busy), 
				Thread_ipc_in_progress | ipc_state))
	{
	  state_del(Thread_busy);
	  continue;
	}

      schedule();
    }

  if (timeout.is_set())
    timeout.reset();

  my_timeout = 0;
  
  while ((state() & (ipc_state | Thread_busy |
		     Thread_ipc_in_progress | Thread_cancel))
	 == (Thread_busy|Thread_ipc_in_progress)) // long IPC?
    {
      // XXX page fault timeouts
      
      if (my_pagein_request)
	{
	  handle_page_fault_pager(my_pagein_request, 2);

	  my_pagein_request = 0;

	  my_partner->ipc_receiver_ready(); // resume IPC
	}

      if (! state_change_safely(~(Thread_ipc_in_progress 
				  | Thread_busy | Thread_running), 
				Thread_ipc_in_progress | Thread_busy))
	{
	  break;
	}

      schedule();
    }


  if (state() & (Thread_busy | ipc_state)) // abnormal termination
    {
      // the IPC has not been finished.  could be timeout or cancel
      state_del(Thread_busy | ipc_state);

      if (state() & Thread_cancel)
	{
	  // we've presumably been reset!
	  regs->eax = L4_IPC_RECANCELED;
	}
      else
	regs->eax = L4_IPC_RETIMEOUT;
    }

  return regs->eax;		// sender puts return code here
}

// the IPC system call
unsigned
thread_t::sys_ipc(thread_regs_t *regs)
{
  unsigned ret = 0;
  l4_timeout_t t;
  t.timeout = regs->ecx;

  // find the ipc partner thread belonging to the destination thread
  threadid_t dst_id(regs->thread_esi_edi());
  thread_t *partner = dst_id.lookup(); // XXX no clans & chiefs for now!

  // first, before starting a send, we must prepare a pending receive
  // so that we can atomically switch from send mode to receive mode
  
  bool have_receive_part = (regs->ebp != 0xffffffff);
  bool have_send_part = (regs->eax != 0xffffffff);
  bool have_sender = false;
  sender_t *sender = 0;
  irq_t *interrupt = 0;
  
  if (have_receive_part)
    {
      if (regs->ebp & 1)	// open wait ?
	{
	  sender = 0; 
	  have_sender = true;
	}

      // not an open wait
      else if (! dst_id.is_nil()) // not nil thread, and not irq id?
	{
	  if (! partner)
	    return L4_IPC_ENOT_EXISTENT;

	  sender = partner; 
	  have_sender = true;
	}

      // nil thread or interrupt thread
      else if (regs->esi == 0)	// nil thread?
	{
	  // only prepare IPC for timeout != 0
	  if (t.to.rcv_exp == 0 || t.to.rcv_man != 0)
	    {
	      sender = partner; 
	      have_sender = true;
	    }
	}
      // must be interrupt thread
      else 
	{
	  if (regs->esi > irq_t::irq_max)
	    return L4_IPC_ENOT_EXISTENT;

	  interrupt = irq_t::lookup(regs->esi - 1);

	  if (my_irq != interrupt)
	    {
	      // a receive with a timeout != 0 from a non-associated
	      // interrupt is illegal
	      if (t.to.rcv_exp == 0 || t.to.rcv_man != 0) // t/o != 0
		return L4_IPC_ENOT_EXISTENT;
	    }

	  if (my_irq)
	    {
	      sender = my_irq;	// we always try to receive from the
				// assoc'd irq first, not from the spec. one
	      have_sender = true;
	    }
	}

      if (have_sender)
	prepare_receive(sender, regs);
    }

  // The send part.  Be careful here: We must not test the parameter
  // in the %eax register here because %eax may already have been
  // overwritten by a sender if this IPC operation has a receive part
  // but no send part (because in this case, prepare_receive() already
  // activates the receive part).  That's why, we save that parameter
  // early (have_send_part).

  if (have_send_part)		// do we do a send operation?
    {
      ret = do_send(partner, t, regs);
    }

  // send done, receive operation follows

  if (have_receive_part		// do we do a receive operation?
      && (ret & L4_IPC_ERROR_MASK) == 0) // no send error
    {
      if (have_sender)
	{
	  // do the receive operation
	  ret = do_receive(sender, t, regs);

	  // if this was a receive from an interrupt and the timeout was 0, 
	  // this is a re-associate operation
	  if (ret != L4_IPC_RETIMEOUT
	      || ! interrupt
	      || (t.to.rcv_exp == 0 || t.to.rcv_man != 0)) // t/o != 0
	    {
	      return ret;
	    }
	}

      else if (! interrupt)	// rcv from nil; this also means t/o == 0
	{
	  //
	  // disassociate from irq
	  //
	  if (my_irq)
	    {
	      my_irq->free(this);
	      my_irq = 0;
	    }

	  return L4_IPC_RETIMEOUT;
	}

      //
      // associate with an interrupt
      //
      if (interrupt->alloc(this))
	{
	  // succeeded -- free previously allocated irq
	  if (my_irq)
	    my_irq->free(this);

	  my_irq = interrupt;

	  return L4_IPC_RETIMEOUT; // success
	}

      return L4_IPC_ENOT_EXISTENT; // failed -- could not associate new irq
    }

  // skipped receive operation
  state_del(Thread_receiving|Thread_waiting|Thread_ipc_in_progress);

  return ret;
}

// the page fault handler
bool 
thread_t::handle_page_fault_pager(vm_offset_t pfa, unsigned error_code)
{
  // do not handle user space page faults from kernel mode if we're
  // already handling a signal
  if (!(error_code & 4) && switching_in.test())
    {
      panic("page fault in signal handler");
    }

  // set up a register block used as an IPC parameter block for the
  // page fault IPC

  thread_user_regs_t r;
  
  r.edx = (pfa & ~3) | (error_code & 3); // dword 1
  r.ebx = (error_code & 4) ? regs()->eip : 0xffffffff; // dword 2
  r.eax = 0;			// snd descriptor = short msg
  r.ebp = reinterpret_cast<unsigned>
    (L4_IPC_MAPMSG(0, L4_WHOLE_ADDRESS_SPACE)); // rcv descriptor = map msg
  
  // This might be a page fault in midst of a long-message IPC operation.
  // Save the current IPC state and restore it later.
  sender_t *orig_partner = my_partner;
  thread_t *orig_send_partner = my_send_partner;
  unsigned orig_ipc_state = state() 
    & (Thread_busy | Thread_polling | Thread_send_in_progress 
       | Thread_ipc_in_progress);
  state_del(orig_ipc_state);

  prepare_receive(my_pager, &r);

  unsigned err = do_send(my_pager, L4_IPC_NEVER, &r);
  bool ret = true;

  if (err & L4_IPC_ERROR_MASK)
    {
      if (config::conservative)
	{
	  printf("page fault send error = 0x%x\n", err);
	  kdb::ke("send to pager failed");
	}

      // skipped receive operation
      state_del(Thread_receiving|Thread_waiting|Thread_ipc_in_progress);

      if ((err & L4_IPC_ERROR_MASK) == L4_IPC_ENOT_EXISTENT)
	ret = (state() & Thread_cancel); //retry user insn after thread_ex_regs
      else
	ret = true;		// retry
    }
  else
    {
      err = do_receive(my_pager, L4_IPC_NEVER, &r);
  
      if (err & L4_IPC_ERROR_MASK)
	{
	  if (config::conservative)
	    {
	      printf("page fault rcv error = 0x%x\n", err);
	      kdb::ke("rcv from pager failed");
	    }
	  
	}

      // XXX currently, we (Jochen-compatibly) ignore receive errors here
      // and just retry the page fault.  However, maybe it would make
      // sense to start exception handling (return false) when the pager
      // returns something which is not a mapping?
      ret = true;
    }

  // restore previous IPC state
  my_partner = orig_partner;
  my_send_partner = orig_send_partner;
  state_add(orig_ipc_state);

  return ret;
}
