#include <l4/sys/types.h>
#include <l4/sys/ipc.h>

#include "config.h"
#include "kmem.h"
#include "thread.h"
#include "ipc.h"

unsigned
thread_t::do_send_long(thread_t *partner, l4_timeout_t t, 
		       thread_user_regs_t *regs)
{
  // just copy the message here.  don't care about pages not being
  // mapped in -- this is completely handled in the page fault handler.
  
  // XXX page fault timeouts not implemented

  // XXX bound check missing!
  
  // set up pointers to message buffers
  
  struct message_header {
    l4_fpage_t   fp;
    l4_msgdope_t size_dope;
    l4_msgdope_t send_dope;
    dword_t      d1, d2;
    dword_t      words[1];
  } *snd_header, *rcv_header;
  
  snd_header = reinterpret_cast<message_header*>(regs->eax & ~3);
  
  // the receiver's message buffer is mapped into VM window 1.
  assert(my_target_desc);
  setup_ipc_window(0, trunc_superpage(my_target_desc));
  rcv_header = reinterpret_cast<message_header*>
    (kmem::ipc_window(0) + (my_target_desc & SUPERPAGE_MASK));
  
  // make sure this is not a register-only send operation -- such
  // operations should have been handled by ipc_send_regs()
  assert(snd_header || (regs->eax & 2));
  
  bool send_fpages = (regs->eax & 2);

  // set up return value in case we're aborted; we must have read
  // all information from eax at this point
  regs->eax = L4_IPC_SEABORTED;
  
  // read the fpage option from the receiver's message buffer
  l4_fpage_t rcv_fpage = rcv_header->fp;

  // check if we cannot send flexpages
  if (send_fpages && ! rcv_fpage.fpage)
    {
      return partner->ipc_finish(L4_IPC_REMSGCUT);
    }

  // send a short flexpage?
  if (send_fpages && !snd_header)
    {
      // OK, read the receiver's flexpage descriptor
      return partner->ipc_send_fpage(((l4_fpage_t){fpage: regs->ebx}),
				     rcv_fpage, 
				     regs->edx & ~PAGE_MASK,
				     true);
    }

  // OK, so we have to send a long message buffer.
  assert(snd_header);

  l4_msgdope_t result = L4_IPC_DOPE(2, 0); // 2 dwords in regs
  unsigned ret = 0;

  // We start by transferring dwords and do flexpage mappings on our way.
  dword_t *cursor = snd_header->words;
  int snd_words = snd_header->send_dope.md.dwords - 2;
  int rcv_words = rcv_header->size_dope.md.dwords - 2;

  if (snd_words < 0) snd_words = 0;
  if (rcv_words < 0) rcv_words = 0;

  unsigned s = snd_words, r = rcv_words;

  unsigned min = (snd_words > rcv_words) ? rcv_words : snd_words;
  if (min)
    {
      memcpy(rcv_header->words, cursor, min * sizeof(dword_t));
      result.md.dwords = min + 2; // 2 dwords in regs
    }

  if (send_fpages)
    {
      while (s >= 2  && r >= 2)
	{
	  l4_fpage_t snd_fpage 
	    = ((l4_fpage_t){fpage: cursor[1]});

	  // if not a valid flexpage, break
	  if (/* snd_fpage.fpage == 0 || */ snd_fpage.fp.size < 12)
	    {
	      send_fpages = false;
	      break;
	    }

	  // otherwise, map
	  ret = partner->ipc_send_fpage(snd_fpage, rcv_fpage, *cursor, false);
	  if (ret & L4_IPC_ERROR_MASK)
	    break;

	  result.msgdope |= ret;

	  cursor += 2;
	  s -= 2;
	  r -= 2;
	}
    }

  if (ret & L4_IPC_ERROR_MASK)
    {      
      // a mapping couldn't be established
      return partner->ipc_finish(ret | result.msgdope);
    }
  else if (snd_words > rcv_words)
    {
      // the message has been cut
      return partner->ipc_finish(L4_IPC_REMSGCUT | result.msgdope);
    }

  // All right -- the direct string in the message buffer has been
  // transferred.  Now tackle the indirect strings.

  l4_strdope_t *snd_strdope = reinterpret_cast<l4_strdope_t *>
    (snd_header->words + (snd_header->size_dope.md.dwords >= 2 
			  ? snd_header->size_dope.md.dwords - 2
			  : 0));
  l4_strdope_t *rcv_strdope = reinterpret_cast<l4_strdope_t *>
    (rcv_header->words + rcv_words);
  int snd_strings = snd_header->send_dope.md.strings;
  int rcv_strings = rcv_header->size_dope.md.strings;

  s = snd_strings;
  r = rcv_strings;

  while (s && r)
    {
      vm_offset_t from = snd_strdope->snd_str;
      vm_offset_t from_size = snd_strdope->snd_size;
      vm_offset_t to = rcv_strdope->rcv_str;
      vm_offset_t to_size = rcv_strdope->rcv_size;

      // silently limit sizes to 4MB
      if (to_size > SUPERPAGE_SIZE) to_size = SUPERPAGE_SIZE;
      if (from_size > SUPERPAGE_SIZE) from_size = SUPERPAGE_SIZE;

      min = (from_size > to_size) ? to_size : from_size;
      if (min > 0)
	{
	  // XXX no bounds checking!
	  setup_ipc_window(1, trunc_superpage(to));
	  memcpy(reinterpret_cast<void*>
		   (kmem::ipc_window(1) + (to & SUPERPAGE_MASK)),
		 reinterpret_cast<void*>(from),
		 min);
	}

      if (from_size > to_size) 
	break;      

      s--;
      r--;
      snd_strdope++;
      rcv_strdope++;

      result.md.strings++;
    }
  
  return partner->ipc_finish(result.msgdope
			     | (s ? L4_IPC_REMSGCUT : 0));

}

bool 
thread_t::handle_ipc_page_fault(vm_offset_t pfa)
{
  vm_offset_t pf_va =  pfa < kmem::ipc_window(1) 
    ? pfa - kmem::ipc_window(0) + my_vm_window0
    : pfa - kmem::ipc_window(1) + my_vm_window1;

  for (;;)
    {
      kernel_lock.lock();
      if (((my_send_partner->state() & (Thread_busy|Thread_ipc_in_progress)) 
	   != (Thread_busy|Thread_ipc_in_progress))
	  || my_send_partner->my_partner != this)
	panic("XXX abort IPC");
      
      if (current_space()->update(pfa, my_send_partner->space(), pf_va));
      {
	kernel_lock.clear();
	return true;
      }
      kernel_lock.clear();
      
      // need to page in data
      my_send_partner->ipc_pagein_request(pf_va);
      
      while (state_change_safely(~(Thread_polling|Thread_ipc_in_progress
				   |Thread_running),
				 Thread_polling|Thread_ipc_in_progress))
	{
	  schedule();
	}
      
      state_add(Thread_polling);
      
      // XXX error handling
    }
}


unsigned
thread_t::ipc_send_fpage(l4_fpage_t from_fpage, l4_fpage_t to_fpage, 
			 vm_offset_t offset, bool finish)
{
  if (! in_transaction())	// not already in signal?
    {
      unsigned ret = trans_invoke(this, &ipc_send_fpage, from_fpage,
				  to_fpage, offset, finish);
      
      // if send_signal() has to abort for some reason, convert to an IPC
      // error code here
      return (ret == static_cast<unsigned long>(signal_t::Sig_aborted))
	? L4_IPC_SECANCELED
	: ret;
    }
  
  if (((state() & (Thread_busy|Thread_ipc_in_progress)) 
       != (Thread_busy|Thread_ipc_in_progress))
      || my_partner != _current_trans->sender())
    return L4_IPC_SEABORTED;

  unsigned ret = fpage_map(_current_trans->sender()->space(), from_fpage,
			   space(), to_fpage,
			   offset);

  if (finish)			// should we finish the IPC operation?
    {
      // if we must finish the IPC operation, set the receiver's return code
      // and return the sender's return code (instead of the receiver's)
      my_receive_regs->eax = ret;
      ret = (ret & L4_IPC_ERROR_MASK) 
	? ((ret & L4_IPC_ERROR_MASK) | L4_IPC_SND_ERR_MASK)
	: 0;

      _current_trans->sender()->state_del(Thread_polling 
					  | Thread_send_in_progress);
      
      state_change(~(Thread_busy | Thread_ipc_in_progress),
		   Thread_running);

      _current_trans->set_switch_hint(signal_t::Ret_activate_receiver);
    }
  else
    _current_trans->set_switch_hint(signal_t::Ret_activate_sender);

  return ret;
}


unsigned 
thread_t::ipc_finish(unsigned new_state)
{
  if (! in_transaction())	// not already in signal?
    {
      unsigned ret = trans_invoke(this, &ipc_finish, new_state);
      
      // if send_signal() has to abort for some reason, convert to an IPC
      // error code here
      return (ret == static_cast<unsigned long>(signal_t::Sig_aborted))
	? L4_IPC_SECANCELED
	: ret;
    }
  
  if (((state() & (Thread_busy|Thread_ipc_in_progress)) 
       != (Thread_busy|Thread_ipc_in_progress))
      || my_partner != _current_trans->sender())
    return L4_IPC_SEABORTED;

  my_receive_regs->eax = new_state;

  _current_trans->sender()->state_del(Thread_polling 
				      | Thread_send_in_progress);
  state_change(~(Thread_busy | Thread_ipc_in_progress),
	       Thread_running);

  _current_trans->set_switch_hint(signal_t::Ret_activate_receiver);

  return (new_state & L4_IPC_ERROR_MASK)
    ? ((new_state & L4_IPC_ERROR_MASK) | L4_IPC_SND_ERR_MASK)
    : 0;
}

unsigned 
thread_t::ipc_pagein_request(vm_offset_t address)
{
  if (! in_transaction())	// not already in signal?
    {
      return trans_invoke(this, &ipc_pagein_request, address);
    }
  
  if (((state() & (Thread_busy|Thread_ipc_in_progress)) 
       != (Thread_busy|Thread_ipc_in_progress))
      || my_partner != _current_trans->sender())
    return signal_t::Sig_fail;

  my_pagein_request = address;

  state_add(Thread_running);

  _current_trans->set_switch_hint(signal_t::Ret_activate_receiver);

  return signal_t::Sig_success;
}
