#include "atomic.h"
#include "thread.h"
#include "thread_list.h"
#include "irq.h"

static irq_t irq00(0);
static irq_t irq01(1);
static irq_t irq02(2);
static irq_t irq03(3);
static irq_t irq04(4);
static irq_t irq05(5);
static irq_t irq06(6);
static irq_t irq07(7);
static irq_t irq08(8);
static irq_t irq09(9);
static irq_t irq10(10);
static irq_t irq11(11);
static irq_t irq12(12);
static irq_t irq13(13);
static irq_t irq14(14);
static irq_t irq15(15);

irq_t * const irq_t::irqs[] = { &irq00, &irq01, &irq02, &irq03, &irq04, 
				&irq05, &irq06, &irq07, &irq08, &irq09, 
				&irq10, &irq11, &irq12, &irq13, &irq14, 
				&irq15 };

// constructor
irq_t::irq_t(unsigned irqnum)
  : _irq_thread(0), _queued(0)
{ 
  my_id = L4_NIL_ID;
  my_id.lh.low = irqnum + 1;

  my_state = Thread_dead;

  // these initializations should be part of sender_t (from which we
  // inherit), but aren't done there with respect to special needs of
  // thread_t

  sender_next = 0;
}

unsigned
thread_t::ipc_send_irq(irq_t *irq)
{
  if (! in_transaction())	// not already in signal?
    {
      return trans_invoke(this, &ipc_send_irq, irq);
    }

  if (state() == Thread_invalid)
    return L4_IPC_ENOT_EXISTENT;

  unsigned ipc_state = state() & (Thread_receiving|Thread_waiting
				  | Thread_send_in_progress
				  | Thread_ipc_in_progress);

  // Thread_send_in_progress must not be set
  if (! (ipc_state == (Thread_waiting | Thread_ipc_in_progress))
      && (! ((ipc_state == (Thread_receiving | Thread_ipc_in_progress)) 
	     && my_partner == irq)))
    return 0x80000000;		// transient error -- partner not ready

  state_change(~(Thread_waiting | Thread_receiving | Thread_busy
		 | Thread_ipc_in_progress),
	       Thread_running);
  // XXX receiver should also get a fresh timeslice

  int o;
  do {
    o = irq->_queued;
  } while (! compare_and_swap(&irq->_queued, o, o - 1));

  if (o <= 1)			// remain queued if more interrupts are left
    irq->sender_dequeue(this);

  *my_receive_regs->thread_esi_edi() = irq->id(); // copy sender ID
  my_receive_regs->eax = 0;	// state = OK

  return 0;			// OK
}

// the receiver's activation function
unsigned 
irq_t::ipc_receiver_ready()
{
  assert(current() == _irq_thread);

  if (_queued)
    _irq_thread->ipc_send_irq(this);
  
  return 0;
}

inline void
irq_t::hit()
{
  // We're entered holding the kernel lock, which also means irqs are
  // disabled on this CPU (XXX always correct?).  We never enable irqs
  // in this stack frame (except maybe in a nonnested invocation of
  // switch_to() -> switchin_context()) -- they will be re-enabled
  // once we return from it (iret in entry.S:all_irqs) or we switch to
  // a different thread.

  if (! _irq_thread)
    {
      pic_disable_irq(my_id.lh.low - 1);
    }

  else if (_queued++ == 0)	// increase hit counter
    {
      my_send_partner = _irq_thread;
      sender_enqueue(_irq_thread); // XXX should enqueue at position 1
      
      // if the thread is waiting for this interrupt, make it running;
      // this will cause it to run irq->receiver_ready(), which
      // handles the rest

      unsigned ipc_state = 
	_irq_thread->state() & (Thread_receiving|Thread_waiting
				| Thread_send_in_progress
				| Thread_ipc_in_progress);
      
      // Thread_send_in_progress must not be set
      if (ipc_state == (Thread_waiting | Thread_ipc_in_progress)
	  || (ipc_state == (Thread_receiving | Thread_ipc_in_progress)
	      && _irq_thread->my_partner == this))
	{
	  // we don't need to manipulate the state in a safe way
	  // because we are still running with interrupts turned off
	  _irq_thread->state_change(~(ipc_state | Thread_busy), 
				    ipc_state | Thread_running);
	  if (current() != _irq_thread
	      && _irq_thread->my_prio >= current()->my_prio)
	    {
	      // kernel-unlock is done in switch_to() (on switchee's side)
	      
	      // switch to the interrupt thread	 
	      current()->switch_to(_irq_thread);
	  
	      return;
	    }
	  else
	    _irq_thread->ready_enqueue();
	}
    }
}

// hardware interrupt entry point
extern "C" void
irq_interrupt(unsigned irqnum)
{
  // we're entered with disabled irqs
#if 0
  irq_ack(irqnum);		// XXX user thread should do this
#endif

#if 0				// screen spinner for debugging purposes
  (*(unsigned char*)(kmem::phys_to_virt(0xb8000 + irqnum*2)))++;
#endif

  kernel_lock.lock();
  irq_t::lookup(irqnum)->hit();
  kernel_lock.clear_irqdisable();
}
