// -*- c++ -*-
#ifndef MAPPING_H
#define MAPPING_H

#include <flux/x86/types.h>
#include <assert.h>

#include "space.h"

enum mapping_type_t { Map_mem = 0, Map_io };

class mapping_tree_t;		// forward decl for private class

class mapping_t
{
public:
  space_index_t space() const;	// the space this mapping belongs to
  vm_offset_t vaddr();		// the virtual address in the space for this
				// mapping
  vm_size_t   size();		// size of the mapping
  mapping_type_t type();	// type of mapping
  mapping_t *parent();		// get parent mapping
  mapping_t *next_iter();	// get child, next sibling, or parent's next
				// sibling
  mapping_t *next_child(mapping_t *parent); // get next anchestor of parent

private:
  friend class mapdb_t;
  friend class mapping_tree_t;

  // default constructors are private: mapping must be created in the
  // mapdb.
  mapping_t() {};		// this constructor actually exists.
  mapping_t(const mapping_t&);	// this constructor is undefined.

  // private data structures
  struct mapping_s
  {
    unsigned space:11;
    unsigned size:1;
    unsigned address:20;
    unsigned depth:8;
    unsigned do_not_touch: 24; // make this 64 bits, and make sure the
                               // compiler never touches memory after the 
                               // real data
  };

  // private functions
  mapping_s *data() const;
  mapping_tree_t *tree() const;
  bool unused() const;

  // data daclaration follows
  char _data[5];
} __attribute__((packed));

//
// implementaion of mapping_t inline methodes
//

inline mapping_t::mapping_s *
mapping_t::data() const
{
  return reinterpret_cast<mapping_s*>(_data);
}

inline space_index_t 
mapping_t::space() const
{
  return space_index_t(data()->space);
};

inline vm_offset_t 
mapping_t::vaddr()
{
  return (data()->address << PAGE_SHIFT);
};

inline vm_size_t 
mapping_t::size()
{
  if ( data()->size ) 
    return SUPERPAGE_SIZE;
  else 
    return PAGE_SIZE;
};

inline mapping_type_t 
mapping_t::type()
{
  // return data()->type;;
  return Map_mem;
};

// 
// The mapping database
// 
class mapdb_t
{
public:
  mapdb_t();

  mapping_t *insert(mapping_t *parent,   // insert a new mapping as a
		    space_t *space,      // "child" of "parent"
		    vm_offset_t va, 
		    vm_size_t size,
		    mapping_type_t type);
  void grant(mapping_t *m, space_t *new_space, vm_offset_t va);
  mapping_t *lookup(space_t *space,
		    vm_offset_t va,
		    mapping_type_t type);
  void free(mapping_t* m);      // free the lock to the tree of this mapping
  bool split(mapping_t *m);	// split a large (4MB) mapping into many
				// small ones
  bool flush(mapping_t *m, bool me_too); // flush a single mapping subtree
};

#endif
