// -*- c++ -*-

#ifndef SIGNAL_H
#define SIGNAL_H

#include <assert.h>
#include "list.h"
#include "globals.h"

// select the fast (but type-unsafe) way or the type-safe (but slower:
// one more function call) way of calling transaction functions
#define TRANS_TYPESAFE

class thread_t;

// signal_t is intended as an abstract base class for actual
// signal classes
class signal_t
{
public:
  enum {
    Sig_aborted = -1,
    Sig_success,
    Sig_fail
  };

  enum {
    Ret_dontcare = 0,
    Ret_activate_receiver,
    Ret_activate_sender
  };

  // status requests
  bool done();			// marks signal as handled
  unsigned long ret();		// return value of sig func;
				// valid if done() == 1
  unsigned switch_hint();

  thread_t *sender();

  // manipulations
  void set_ret(unsigned long ret); // also sets done(); called by scheduler
				// when signal is done
  void set_switch_hint(unsigned hint);

#ifdef TRANS_TYPESAFE
  // signal function -- to be overloaded
  virtual unsigned long func(thread_t *target) = 0;

  // constructor
  signal_t();
#else
  unsigned long func(thread_t *target);

  // constructor
  signal_t(void (*func)(...), 
	   void *args, size_t size);
#endif

private:
  signal_t(signal_t&); // default constructors are undefined

  // implementation details follow...

  // data layout 
  union {
    struct {
      // thread ptr's lowest bits are always 0, that's why we can use
      // them for other stuff
      unsigned switch_back: 2;
      unsigned done : 1;
      unsigned other : 29;
    } my_state;
    unsigned long my_sender;
  };

  union {
    unsigned long my_ret;	// return value of sig function...
    sllist_elem_t<signal_t> list_property; // ...or its list property (next pointer)
  };

#ifndef TRANS_TYPESAFE
  void (*_func)(...);
  void *_args;
  size_t _size;
#endif

  friend class list_stack_t<signal_t>; // so that we can be enqueued
};


//
// inline stuff for signal_t
//

#include "thread.h"

inline bool signal_t::done()
{
  return my_state.done;
}

inline unsigned long signal_t::ret()
{
  assert(done());
  return my_ret;
}

inline thread_t *signal_t::sender()
{
  return reinterpret_cast<thread_t*>(my_sender & ~(THREAD_BLOCK_SIZE - 1));
}

inline unsigned signal_t::switch_hint()
{
  return my_state.switch_back;
}

inline void signal_t::set_ret(unsigned long ret)
{
  my_ret = ret;
  my_state.done = 1;
}

inline void signal_t::set_switch_hint(unsigned hint)
{
  my_state.switch_back = hint;
}

#ifdef TRANS_TYPESAFE

inline signal_t::signal_t()
{
  my_sender = reinterpret_cast<unsigned long>(current());
}

class trans_signal0_t : public signal_t
{
  unsigned (thread_t::*_func)();

  virtual unsigned long func(thread_t *target)
    {
      return (target->*_func)();
    }

public:
  trans_signal0_t(unsigned (thread_t::*func)())
    : signal_t(),
      _func(func)
  {}
};

inline unsigned 
trans_invoke(thread_t *target,
	     unsigned (thread_t::*func)())
{
  trans_signal0_t s(func);

  return current()->send_signal(target, &s);
}

template<class T1>
class trans_signal1_t : public signal_t
{
  T1 _v1;

  unsigned (thread_t::*_func)(T1 v1);

  virtual unsigned long func(thread_t *target)
    {
      return (target->*_func)(_v1);
    }

public:
  trans_signal1_t(unsigned (thread_t::*func)(T1 v1),
		  T1 v1)
    : signal_t(),
      _v1(v1), _func(func)
  {}
};

template<class T1>
inline unsigned 
trans_invoke(thread_t *target,
	     unsigned (thread_t::*func)(T1 v1),
	     T1 v1)
{
  trans_signal1_t<T1> s(func, v1);

  return current()->send_signal(target, &s);
}

template<class T1, class T2>
class trans_signal2_t : public signal_t
{
  T1 _v1;
  T2 _v2;

  unsigned (thread_t::*_func)(T1 v1, T2 v2);

  virtual unsigned long func(thread_t *target)
    {
      return (target->*_func)(_v1, _v2);
    }

public:
  trans_signal2_t(unsigned (thread_t::*func)(T1 v1, T2 v2),
		  T1 v1, T2 v2)
    : signal_t(),
      _v1(v1), _v2(v2), _func(func)
  {}
};

template<class T1, class T2>
inline unsigned 
trans_invoke(thread_t *target,
	     unsigned (thread_t::*func)(T1 v1, T2 v2),
	     T1 v1, T2 v2)
{
  trans_signal2_t<T1,T2> s(func, v1, v2);

  return current()->send_signal(target, &s);
}

template<class T1, class T2, class T3, class T4>
class trans_signal4_t : public signal_t
{
  T1 _v1;
  T2 _v2;
  T3 _v3;
  T4 _v4;

  unsigned (thread_t::*_func)(T1 v1, T2 v2, T3 v3, T4 v4);

  virtual unsigned long func(thread_t *target)
    {
      return (target->*_func)(_v1, _v2, _v3, _v4);
    }

public:
  trans_signal4_t(unsigned (thread_t::*func)(T1 v1, T2 v2, T3 v3, T4 v4),
		 T1 v1, T2 v2, T3 v3, T4 v4)
    : signal_t(),
      _v1(v1), _v2(v2), _v3(v3), _v4(v4), _func(func)
  {}
};

template<class T1, class T2, class T3, class T4>
inline unsigned 
trans_invoke(thread_t *target,
	     unsigned (thread_t::*func)(T1 v1, T2 v2, T3 v3, T4 v4),
	     T1 v1, T2 v2, T3 v3, T4 v4)
{
  trans_signal4_t<T1,T2,T3,T4> s(func, v1, v2, v3, v4);

  return current()->send_signal(target, &s);
}

template<class T1, class T2, class T3, class T4, 
  class T5, class T6>
class trans_signal6_t : public signal_t
{
  T1 _v1;
  T2 _v2;
  T3 _v3;
  T4 _v4;
  T5 _v5;
  T6 _v6;

  unsigned (thread_t::*_func)(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, 
			      T6 v6);

  virtual unsigned long func(thread_t *target)
    {
      return (target->*_func)(_v1, _v2, _v3, _v4, _v5, _v6);
    }

public:
  trans_signal6_t(unsigned (thread_t::*func)(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5,
					     T6 v6),
		 T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, T6 v6)
    : signal_t(),
      _v1(v1), _v2(v2), _v3(v3), _v4(v4), _v5(v5), _v6(v6), _func(func)
  {}
};

template<class T1, class T2, class T3, class T4, 
  class T5, class T6>
inline unsigned 
trans_invoke(thread_t *target,
	     unsigned (thread_t::*func)(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, 
					T6 v6),
	     T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, T6 v6)
{
  trans_signal6_t<T1,T2,T3,T4,T5,T6> 
    s(func, v1, v2, v3, v4, v5, v6);

  return current()->send_signal(target, &s);
}

template<class T1, class T2, class T3, class T4, 
  class T5, class T6, class T7, class T8, class T9>
class trans_signal9_t : public signal_t
{
  T1 _v1;
  T2 _v2;
  T3 _v3;
  T4 _v4;
  T5 _v5;
  T6 _v6;
  T7 _v7;
  T8 _v8;
  T9 _v9;

  unsigned (thread_t::*_func)(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, 
			      T6 v6, T7 v7, T8 v8, T9 v9);

  virtual unsigned long func(thread_t *target)
    {
      return (target->*_func)(_v1, _v2, _v3, _v4, _v5, _v6, _v7, _v8, _v9);
    }

public:
  trans_signal9_t(unsigned (thread_t::*func)(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5,
					     T6 v6, T7 v7, T8 v8, T9 v9),
		 T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, T6 v6, T7 v7, T8 v8, T9 v9)
    : signal_t(),
      _v1(v1), _v2(v2), _v3(v3), _v4(v4), _v5(v5), _v6(v6), _v7(v7), _v8(v8), 
      _v9(v9), _func(func)
  {}
};

template<class T1, class T2, class T3, class T4, 
  class T5, class T6, class T7, class T8, class T9>
inline unsigned 
trans_invoke(thread_t *target,
	     unsigned (thread_t::*func)(T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, 
					T6 v6, T7 v7, T8 v8, T9 v9),
	     T1 v1, T2 v2, T3 v3, T4 v4, T5 v5, T6 v6, T7 v7, T8 v8, T9 v9)
{
  trans_signal9_t<T1,T2,T3,T4,T5,T6,T7,T8,T9> 
    s(func, v1, v2, v3, v4, v5, v6, v7, v8, v9);

  return current()->send_signal(target, &s);
}


#else  // ! TRANS_TYPESAFE

inline 
signal_t::signal_t(void (*func)(...), 
		   void *args, size_t size)
  : _func(func), _args(args), _size(size)
{
  my_sender = reinterpret_cast<unsigned long>(current());
}

// help EGCS: the internal declaration of __builtin_apply() is wrong
#if __GNUC__ == 2 && __GNUC_MINOR__ > 8	
extern "C" void * __builtin_apply(void (*func)(...), 
				  void *args, size_t size);
#endif

inline unsigned long
signal_t::func(thread_t * /*target*/ )
{
  __builtin_return(__builtin_apply(_func, _args, _size));
}

#define trans_invoke(target, func, args...)			\
({								\
  signal_t s((void (*)(...))(func),				\
	     __builtin_apply_args(), trans_sizeofargs(args));	\
  current()->send_signal((target), &s);				\
})

inline unsigned trans_sizeofargs()
{
  return sizeof(void*);
}

template <class T1>
inline unsigned trans_sizeofargs(T1 a1)
{
  struct foo {
    T1 v1;
  };

  return sizeof(foo) + sizeof(void*);
}

template <class T1, class T2> 
inline unsigned trans_sizeofargs(T1 a1, T2 a2)
{
  struct foo {
    T1 v1;
    T2 v2;
  };

  return sizeof(foo) + sizeof(void*);
}

template <class T1, class T2, class T3> 
inline unsigned trans_sizeofargs(T1 a1, T2 a2, T3 a3)
{
  struct foo {
    T1 v1;
    T2 v2;
    T3 v3;
  };

  return sizeof(foo) + sizeof(void*);
}

template <class T1, class T2, class T3, 
  class T4> 
inline unsigned trans_sizeofargs(T1 a1, T2 a2, T3 a3, T4 a4)
{
  struct foo {
    T1 v1;
    T2 v2;
    T3 v3;
    T4 v4;
  };

  return sizeof(foo) + sizeof(void*);
}

template <class T1, class T2, class T3, 
  class T4, class T5> 
inline unsigned trans_sizeofargs(T1 a1, T2 a2, T3 a3, T4 a4, 
				 T5 a5)
{
  struct foo {
    T1 v1;
    T2 v2;
    T3 v3;
    T4 v4;
    T5 v5;
  };

  return sizeof(foo) + sizeof(void*);
}

template <class T1, class T2, class T3, 
  class T4, class T5, class T6> 
inline unsigned trans_sizeofargs(T1 a1, T2 a2, T3 a3, T4 a4, 
				 T5 a5, T6 a6)
{
  struct foo {
    T1 v1;
    T2 v2;
    T3 v3;
    T4 v4;
    T5 v5;
    T6 v6;
  };

  return sizeof(foo) + sizeof(void*);
}

template <class T1, class T2, class T3, 
  class T4, class T5, class T6,
  class T7, class T8> 
inline unsigned trans_sizeofargs(T1 a1, T2 a2, T3 a3, T4 a4, 
				 T5 a5, T6 a6, T7 a7, 
				 T8 a8)
{
  struct foo {
    T1 v1;
    T2 v2;
    T3 v3;
    T4 v4;
    T5 v5;
    T6 v6;
    T7 v7;
    T8 v8;
  };

  return sizeof(foo) + sizeof(void*);
}

template <class T1, class T2, class T3, class T4, class T5, class T6,
  class T7, class T8, class T9> 
inline unsigned trans_sizeofargs(T1 a1, T2 a2, T3 a3, T4 a4, 
				 T5 a5, T6 a6, T7 a7, 
				 T8 a8, T9 a9)
{
  struct foo {
    T1 v1;
    T2 v2;
    T3 v3;
    T4 v4;
    T5 v5;
    T6 v6;
    T7 v7;
    T8 v8;
    T9 v9;
  };

  return sizeof(foo) + sizeof(void*);
}

#endif // ! TRANS_TYPESAFE

#endif // ! SIGNAL_H
