#include <stdlib.h>
#include <stdio.h>

#include <oskit/machine/proc_reg.h>
#include <oskit/machine/pio.h>
#include <oskit/machine/pc/pic.h>
#include <oskit/dev/dev.h>

#include <l4/compiler.h>
#include <l4/sys/types.h>
#include <l4/sys/syscalls.h>
#include <l4/sys/ipc.h>
#include <l4/rmgr/librmgr.h>

#include <l4/oskit10/support.h>

/* hack: symbols that can be referenced to suck in this object */
asm("l4_oskit_support_osenv_irq:");
asm("l4_oskit_support_osenv_intr:");

#define IRQ_THREAD_STACK_SIZE (512)
#define MANUAL_IRQ_ACK
static unsigned irq_lthread_no_offset = 0x40;

static unsigned irq_pending = 0, irq_blocked = ~0;
static unsigned irq_thread_started = 0;
static unsigned intr_enabled = 1;

struct irqaction
{
  void (*handler)(void *);
  void *data;
  int flags;
  struct irqaction *next;
};

static struct irqaction *irq_handler[NINTR];

static l4_threadid_t irq_driver_thread = L4_NIL_ID,
  irq_driver_pager, irq_driver_preempter;

/* helpers */

static void deliver_irq(int irq);
static void irq_thread(int irq);
L4_INLINE int compare_and_swap(unsigned *ptr, unsigned oldval,
			       unsigned newval);

L4_INLINE int
compare_and_swap(unsigned *ptr, unsigned oldval,
                 unsigned newval)
{
  char ret;

  asm volatile
    ("cmpxchgl %1,(%2) ; sete %%al"
     : "=a" (ret)
     : "q" (newval), "q" (ptr), "0" (oldval)
     : "memory");

  return ret;
}

/* implementation of the oskit_{irq,intr} interfaces */

void
osenv_intr_disable(void)
{
  intr_enabled = 0;
}

void
osenv_intr_enable(void)
{
  intr_enabled = 1;
  deliver_irq(-1);
}

int 
osenv_intr_enabled(void)
{
  return intr_enabled;
}

void
osenv_irq_enable(int irq)
{
  irq_blocked &= ~(1 << irq);
  deliver_irq(irq);
}

void
osenv_irq_disable(int irq)
{
  irq_blocked |= (1 << irq);
}

int 
osenv_irq_pending(int irq)
{
  return (irq_pending & (1 << irq)) ? 1 : 0;
}

int 
osenv_irq_alloc(int irq, void (*handler)(void*), void *data, int flags)
{
  struct irqaction *a;
  int ret = -1;
  int was_enabled = osenv_intr_enabled();

  osenv_intr_disable();

  if (irq_handler[irq] 
      && (!(flags & OSENV_IRQ_SHAREABLE)
	  || !(irq_handler[irq]->flags & OSENV_IRQ_SHAREABLE)))
    goto out;

  if (! ((1 << irq) & irq_thread_started))
    {
      /* need to start irq thread */

      unsigned long *stack;
      l4_threadid_t irq_thread_id, preempter, pager;
      dword_t dummy, code, ipcerror;
      l4_msgdope_t result;

      if (thread_equal(irq_driver_thread, L4_NIL_ID))
	{
	  irq_driver_thread = l4_myself();
	  irq_driver_pager = irq_driver_preempter = L4_INVALID_ID;
	  l4_thread_ex_regs(irq_driver_thread, (dword_t) -1, (dword_t) -1,
			    &irq_driver_preempter,
			    &irq_driver_pager,
			    &dummy, &dummy, &dummy);
	}


      stack = malloc(IRQ_THREAD_STACK_SIZE);
      if (! stack)
	goto out;

      stack += IRQ_THREAD_STACK_SIZE/sizeof(*stack);

      /* call irq_thread with args */
      *--stack = irq;
      *--stack = 0;		/* fake return address */

      irq_thread_id = irq_driver_thread;
      irq_thread_id.id.lthread = irq_lthread_no_offset + irq;
      preempter = irq_driver_preempter;
      pager = irq_driver_pager;
      l4_thread_ex_regs(irq_thread_id, (dword_t) irq_thread, (dword_t) stack,
			&preempter, &pager, &dummy, &dummy, &dummy);
      
      /* wait for handshake IPC */
      do 
	{
	  ipcerror = l4_i386_ipc_receive(irq_thread_id, 0, &code, &dummy, 
					 L4_IPC_NEVER, &result);
	}
      while (ipcerror == L4_IPC_RECANCELED);

      if (ipcerror || code != 1)
	goto out;

      irq_thread_started |= (1 << irq);
    }

  a = malloc(sizeof(*a));
  if (!a)
    goto out;

  a->handler = handler;
  a->data = data;
  a->flags = flags;
  a->next = irq_handler[irq];

  irq_handler[irq] = a;

  osenv_irq_enable(irq);

  ret = 0;			/* Success */

 out:
  if (was_enabled) 
    osenv_intr_enable();

  return ret;
}

void 
osenv_irq_free(int irq, void (*handler)(void*), void *data)
{
  struct irqaction *a, *prev = 0;
  int was_enabled = osenv_intr_enabled();

  osenv_intr_disable();

  for (a = irq_handler[irq]; a; a = a->next)
    {
      if (a->handler == handler
	  && a->data == data)
	{
	  if (prev)
	    prev->next = a->next;
	  else
	    irq_handler[irq] = a->next;

	  free(a);
	  break;
	}

      prev = a;
    }

  if (was_enabled) 
    osenv_intr_enable();
}

/* helpers */

void 
oskit_support_irq_init(unsigned lthread_no_offset)
{
  irq_lthread_no_offset = lthread_no_offset;
}

void do_nothing_loop(void);
asm("do_nothing_loop: jmp do_nothing_loop");

void wrap_deliver_signal(int irq);
asm("
    .p2align 2
wrap_deliver_signal:
    pushl %eax
    pushl %ecx
    pushl %edx
    movl  12(%esp),%eax
    pushl %eax
    call  deliver_irq
    leal  4(%esp), %esp
    popl  %edx
    popl  %ecx
    popl  %eax
    leal  4(%esp), %esp
    popfl
    ret");

static void 
irq_thread(int irq)
{
  l4_threadid_t irq_th, pager, preempter;
  dword_t dummy, d1, eip, eflags, *esp;
  l4_msgdope_t dummydope;
  int code;

  /* first, register to the irq */

  /* in case we're running under a RMGR, fetch the irq nr first from it */
  d1 = rmgr_get_irq(irq);
  if (d1 != 0)
    {
      /* the supervisor thread exists and sent a negative response */
      printf("irq_thread: RMGR denied irq %u: code 0x%x\n", irq, d1);
      goto error;
    }

  irq_th.lh.low = irq + 1;
  irq_th.lh.high = 0;

  code = l4_i386_ipc_receive(irq_th,
			     0, /* receive descriptor */
			     &dummy,
			     &dummy,
			     L4_IPC_TIMEOUT(0,0,0,1,0,0), /* rcv = 0,
							     snd = inf */
			     &dummydope);
  if (code != L4_IPC_RETIMEOUT)
    {
      printf("irq_thread: can't register to irq %u: error 0x%x\n", irq, 
	     (unsigned) code);

    error:
      l4_i386_ipc_send(irq_driver_thread, 
		       0, 0, 0,
		       L4_IPC_NEVER, &dummydope);

      /* sleep forever */
      l4_i386_ipc_receive(L4_NIL_ID, 0,
			  &dummy, &dummy,
			  L4_IPC_NEVER,
			  &dummydope);
    }

  /* synchronize with the thread that started us */
  l4_i386_ipc_send(irq_driver_thread, 
		   0, 1, 0,
		   L4_IPC_NEVER, &dummydope);

  /* OK, start handling irq messages */
  for(;;)
    {
      /* this code will eventually go away when L4 will unblock/acknowledge
	 irqs itself */

#ifdef MANUAL_IRQ_ACK
      unsigned char mask;

      cli();			/* run with disabled irqs to avoid races
				   when we acknowledge the interrupt later */
      mask = ~(1 << (irq & 7));
      if (irq < 8) 
	outb(0x21, inb(0x21) & mask);
      else
	outb(0xA1, inb(0xA1) & mask);
#endif /* MANUAL_IRQ_ACK */

      for (;;)
	{
	  code = l4_i386_ipc_receive(irq_th,
				     0, /* receive descriptor */
				     &dummy,
				     &dummy,
				     L4_IPC_NEVER,
				     &dummydope);
	  if (code == 0)
	    break;
	  
	  printf("irq_thread: irq %u (%x:%x) receive failed, code = 0x%x", 
		 irq, irq_th.lh.high, irq_th.lh.low,
		 (unsigned) code);
	}

#ifdef MANUAL_IRQ_ACK
      /* this code will eventually go away when L4 will acknowledge
	 irqs itself */
      
      /* we assume we're still cli'd here; we definitely don't want
	 any other L4 task to manipulate the PIC at the same time */
      
      mask = 1 << (irq & 7);
      if (irq < 8)
	{
	  outb(0x21, inb(0x21) | mask);	/* block the irq */
	  outb(0x20, 0x20);	/* acknowledge the irq */
	}
      else 
	{
	  outb(0xA1, inb(0xA1) | mask);	/* block */
	  outb(0xA0, 0x20);	/* acknowledge */
	  outb(0xA0, 0x0B);
	  if (inb(0xA0) == 0)
	    outb(0x20, 0x20);
	}      
  
      sti();
#endif /* MANUAL_IRQ_ACK */

      irq_pending |= (1 << irq);

      if ((irq_blocked & (1 << irq))
	  || ! intr_enabled)
	{
	  continue;		/* blocked -- don't deliver */
	}

      /* deliver signal to handler thread */
      pager = preempter = L4_INVALID_ID;
      l4_thread_ex_regs(irq_driver_thread, (dword_t) do_nothing_loop, 0,
			&preempter, &pager, &eflags, &eip, (dword_t*) &esp);

      /* someone else modifying state already? */
      if (eip == (dword_t) do_nothing_loop)
	continue;		/* do not interfere */

      /* push return frame */
      *--esp = eip;
      *--esp = eflags;

      /* fake invokation of wrap_deliver_signal */
      *--esp = irq;
      l4_thread_ex_regs(irq_driver_thread, (dword_t) wrap_deliver_signal, 
			(dword_t) esp,
			&preempter, &pager, &dummy, &dummy, &dummy);


      /* that's it! */
    }
}

static void 
call_handlers(int irq)
{
  struct irqaction *a;
  unsigned old;
  
  do {
    old = irq_pending;
  } while (! compare_and_swap(&irq_pending, old, old & ~(1<<irq)));
  
  for (a = irq_handler[irq]; a; a = a->next)
    {
      a->handler(a->data);
    }
}


static void 
deliver_irq(int irq)
{
  if (! compare_and_swap(&intr_enabled, 1, 0))
    return;
  
  if (irq > 0)
    {
      if (irq_pending & ~irq_blocked & (1 << irq))
	call_handlers(irq);
    }
  
  for (;;)
    {
      if (irq_pending & ~irq_blocked)
	{
	  for (irq = 0; irq < NINTR; irq++)
	    {
	      if (irq_pending & ~irq_blocked & (1 << irq))
		call_handlers(irq);
	    }
	}	  

      intr_enabled = 1;

      if (! (irq_pending & ~irq_blocked))
	return;

      if (! compare_and_swap(&intr_enabled, 1, 0))
	return;
    }
}
