/*
 * Copyright (c) 1996, 1998, 1999 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */
/*
 * Default implementation of memory routines.
 */
#include <oskit/c/malloc.h>
#include <oskit/c/stdlib.h>
#include <oskit/dev/dev.h>
#include <oskit/com/mem.h>
#include <oskit/com/services.h>
#include <oskit/x86/base_vm.h>
#include <oskit/x86/paging.h>
#include <oskit/x86/pc/phys_lmm.h>
#include <oskit/amm.h>

#include "sigma0.h"

/*
 * Stash a pointer to the memory object.
 */
static oskit_mem_t *osenv_mem;

/* keeps track of free regions in VA */
static amm_t address_space;

/*
 * Allocate memory.
 */
void *
osenv_mem_alloc(oskit_size_t size, osenv_memflags_t flags, unsigned align)
{
  void *p;

  if (osenv_mem == NULL) {
    if (oskit_lookup_first(&oskit_mem_iid, (void **)&osenv_mem))
      return NULL;
    if (osenv_mem == NULL)
      return NULL;
  }

  /*
   * NOTE that osenv flags are the same as oskit_mem flags! 
   */

  if (align > 1) {
    p = oskit_mem_alloc_aligned(osenv_mem, size, flags, align);
  } else {
    p = oskit_mem_alloc(osenv_mem, size, flags);
  }

  return p;
}

/*
 * Free memory.
 */
void
osenv_mem_free(void *block, osenv_memflags_t flags, oskit_size_t size)
{
  oskit_mem_free(osenv_mem, block, size, flags);
}

/*
 * Map physical memory into kernel virtual space.
 */
int
osenv_mem_map_phys(oskit_addr_t pa, oskit_size_t size, void **addr,
		   int flags)
{
  oskit_addr_t address, real_size, from, to;

  /* not a hi-mem device page?  just use it! */
  if (pa < phys_mem_max)
    {
      *addr = (void *)phystokv(pa);
      return 0;
    }

  /* first, select a virtual region to hold the I/O mapping */
  real_size = round_superpage(pa + size) - trunc_superpage(pa);
  address = 0;

  if (!amm_find_gen(&address_space, &address, real_size, AMM_FREE, -1,
		    SUPERPAGE_SHIFT, 0, 0))
    return -1;

  if (amm_modify(&address_space, address, real_size, AMM_ALLOCATED, 0) != 0)
    return -1;

  for (to = address, 
	 from = pa & ~SUPERPAGE_MASK;
       from < pa + size;
       to += SUPERPAGE_SIZE, 
	 from += SUPERPAGE_SIZE)
    {
      if (sigma0_request_dev_page(to, from) != RP_4MB)
	{
	  amm_deallocate(&address_space, address, real_size);
	  return -1;
	}
    }

  *addr = (void*) (address + (pa & SUPERPAGE_MASK));
  return 0;
}


/*
 * Return the amount of physical RAM in the machine
 */
oskit_addr_t
osenv_mem_phys_max(void)
{
  return phys_mem_max;
}

/*
 * Return the physical address
 */
oskit_addr_t
osenv_mem_get_phys(oskit_addr_t addr)
{
  return ((oskit_addr_t)kvtophys(addr));
}

oskit_addr_t
osenv_mem_get_virt(oskit_addr_t addr)
{
  return ((oskit_addr_t)phystokv(addr));
}

/* helpers */

void osenv_mem_INIT(void);	/* not declared in a header, just in
				   osenv_mem.c */

void 
osenv_mem_INIT(void)
{
  amm_init(&address_space, round_superpage(phys_mem_max), 0x80000000);
}
